import torch.nn as nn
import torch
from torch.nn import Conv3d, LayerNorm, Parameter, Dropout, Linear, Softmax, ReLU, GroupNorm, Upsample, MaxPool3d
import math


class Embedding(nn.Module):
    def __init__(self, configs):
        super(Embedding, self).__init__()
        d, h, w = configs['input_size']
        patch_size = configs['patch_size']
        hidden_dim = configs['hidden_dim']
        n_patches = (d // patch_size[0]) * (h // patch_size[1]) * (w // patch_size[2])
        self.patch_embbeddings = Conv3d(in_channels=configs['emb_in_channels'],
                                        out_channels=hidden_dim,
                                        kernel_size=patch_size, stride=patch_size)
        self.position_embeddings = Parameter(torch.zeros(1, n_patches, hidden_dim))
        self.dropout = Dropout(configs['embedding_dprate'])

    def forward(self, x):
        x = self.patch_embeddings(x)
        x = x.flatten(2)  # start from 3
        embeddings = x + self.position_embeddings
        embeddings = self.dropout(embeddings)
        return embeddings


class AttentionBlock(nn.Module):
    def __init__(self, hidden_ch, head_num):
        super(AttentionBlock, self).__init__()
        self.head_num = head_num
        self.head_size = hidden_ch // head_num
        self.all_head_size = self.head_num * self.head_size
        self.query = Linear(hidden_ch, self.all_head_size)
        self.key = Linear(hidden_ch, self.all_head_size)
        self.value = Linear(hidden_ch, self.all_head_size)
        self.out = Linear(hidden_ch, hidden_ch)
        self.softmax = Softmax(dim=-1)

    def reshape_multi_heads(self, x):
        new_size = x.size()[:-1] + (self.head_num, self.head_size)
        x = x.view(*new_size)
        return x.permute(0, 2, 1, 3)

    def forward(self, x):
        query = self.query(x)
        key = self.key(x)
        value = self.value(x)

        query_layer = self.reshape_multi_heads(query)
        key_layer = self.reshape_multi_heads(key)
        value_layer = self.reshape_multi_heads(value)

        att_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        att_scores = att_scores / math.sqrt(self.head_size)
        att_probs = self.softmax(att_scores)
        context_layer = torch.matmul(att_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_shape = context_layer.size()[:2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_shape)
        att_out = self.out(context_layer)
        return att_out


class MLP(nn.Module):
    def __init__(self, hidden_ch, mlp_dim, dp_rate):
        super(MLP, self).__init__()
        self.fc1 = Linear(hidden_ch, mlp_dim)
        self.fc2 = Linear(mlp_dim, hidden_ch)
        self.act_fn = nn.functional.gelu
        self.dropout = Dropout(dp_rate)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act_fn(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = self.act_fn(x)
        x = self.dropout(x)
        return x


class TransformerBlock(nn.Module):
    def __init__(self, configs):
        super(TransformerBlock, self).__init__()
        self.att_norm = LayerNorm(configs['hidden_dim'], eps=1e-6)
        self.attention = AttentionBlock(configs['hidden_dim'], configs['head_num'])
        self.mlp = MLP(configs['hidden_dim'], configs['mlp_din'], configs['mlp_dprate'])
        self.fn_norm = LayerNorm(configs['hidden_dim'], eps=1e-6)

    def forward(self, x):
        h = x
        x = self.att_norm(x)
        attention = self.attention(x)
        x = h + attention

        h = x
        x = self.fn_norm(x)
        x = self.mlp(x)
        x = x + h
        return x


class Transformer(nn.Module):
    def __init__(self, configs):
        super(Transformer, self).__init__()
        self.layer = nn.ModuleList()
        self.embeddings = Embedding(configs)
        self.encoder_norm = LayerNorm(configs['hidden_dim'], eps=1e-6)
        for _ in range(configs['trans_layers']):
            transblock = TransformerBlock(configs)
            self.layer.append(transblock)

    def forward(self, x):
        x = self.embeddings(x)
        for layer_block in self.layer:
            x = layer_block(x)
        encoded = self.encoder_norm(x)
        return encoded


class ConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size):
        super(ConvBlock, self).__init__()
        self.block = nn.Sequential(
            Conv3d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, padding='same'),
            ReLU(),
            GroupNorm(num_groups=4, num_channels=out_channels)
        )

    def forward(self, x):
        x = self.block(x)
        return x


class ConvLayer(nn.Module):
    def __init__(self, conv_num, in_channels, out_channels, kernel_size):
        super(ConvLayer, self).__init__()
        self.layer = nn.ModuleList()
        self.block0 = ConvBlock(in_channels, out_channels, 1)
        self.block1 = ConvBlock(in_channels, out_channels, kernel_size)
        for _ in range(conv_num-1):
            self.layer.append(ConvBlock(out_channels, out_channels, kernel_size))

    def forward(self, x):
        h = self.block0(x)
        for layer_block in self.layer:
            x = layer_block(x)
        return h + x


class Encoder(nn.Module):
    def __init__(self, configs):
        super(Encoder, self).__init__()
        self.conv_l1 = ConvLayer(2, 1, 8, 3)
        self.d1 = MaxPool3d(kernel_size=2, stride=2)
        self.conv_l2 = ConvLayer(2, 8, 16, 3)
        self.d2 = MaxPool3d(kernel_size=2, stride=2)
        self.conv_l3 = ConvLayer(2, 16, 32, 3)
        self.transformer = Transformer(configs)

    def forward(self, x):
        cl1 = self.conv_l1(x)
        dl1 = self.d1(cl1)
        cl2 = self.conv_l2(dl1)
        dl2 = self.d2(cl2)
        cl3 = self.conv_l3(dl2)
        out = self.transformer(cl3)
        features = [cl3, cl2, cl1]
        return out, features


class DecoderBlock(nn.Module):
    def __init__(self, in_ch, out_ch, skip_ch):
        super(DecoderBlock, self).__init__()
        self.up = Upsample(scale_factor=2)
        self.downsample = MaxPool3d(2, 2)
        self.conv1 = ConvBlock(in_ch+skip_ch, out_ch, 3)
        self.conv2 = ConvBlock(out_ch, out_ch, 3)

    def forward(self, x, feat):
        x = self.up(x)
        if feat is not None:
            feat = self.downsample(feat)
            x = torch.cat([x, feat], dim=1)
        x = self.conv1(x)
        x = self.conv2(x)
        return x


class CUP(nn.Module):
    def __init__(self, configs):
        super(CUP, self).__init__()
        self.n_skip = configs['n_skip']
        self.patch_size = configs['patch_size']
        d, h, w = configs['input_size']
        patch_size = configs['patch_size']
        self.ori_size = d//patch_size[0], h//patch_size[1], w//patch_size[2]
        decoder_blocks = [DecoderBlock(in_ch, out_ch, skip_ch) for in_ch, out_ch, skip_ch in configs['decoder_channels']]
        self.blocks = nn.ModuleList(decoder_blocks)
        self.conv = ConvBlock(configs['hidden_dim'], configs['head_channels'], 3)

    def forward(self, x, features):
        B, n_patch, hidden_dim = x.size()
        x = x.permute(0, 2, 1)
        new_shape = (B, hidden_dim,) + self.ori_size
        x = x.contiguous().view(*new_shape)
        x = self.conv(x)
        for i, decoder_block in enumerate(self.blocks):
            if features is not None:
                feat = features[i] if i < self.n_skip else None
            else:
                feat = None
            x = decoder_block(x, feat)
        return x


class SegmentationHead(nn.Module):
    def __init__(self, in_ch, n_class):
        super(SegmentationHead, self).__init__()
        self.up = Upsample(scale_factor=2)
        self.head = Conv3d(in_channels=in_ch, out_channels=n_class, kernel_size=1, stride=1)
        self.softmax = Softmax(dim=1)

    def forward(self, x):
        x = self.up(x)
        x = self.head(x)
        x = self.softmax(x)
        return x


class TransUnet(nn.Module):
    def __init__(self, configs):
        super(TransUnet, self).__init__()
        self.encoder = Encoder(configs)
        self.decoder = CUP(configs)
        self.head = SegmentationHead(configs['last_in_ch'], configs['n_class'])

    def forward(self, x):
        x, features = self.encoder(x)
        x = self.decoder(x, features)
        x = self.head(x)
        return x


if __name__ == '__main__':
    img = torch.randn(1, 1, 128, 256, 256)
    configs = {
        'hidden_dim': 256,
        'input_size': (32, 64, 64),
        'patch_size': (4, 4, 4),
        'trans_layers': 2,
        'mlp_dim': 512,
        'embedding_dprate': 0.1,
        'head_num': 4,
        'emb_in_channels': 32,
        'mlp_dprate': 0.2,
        'n_class': 2,
        'last_in_ch': 16,
        'n_skip': 3,
        'head_channels': 128,
        'decoder_channels': [(128, 64, 32), (64, 32, 16), (32, 16, 8)]
    }
    model = TransUnet(configs)
    pred = model(img)
    print(pred.shape)


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is applied to image reading and display.
"""
import cv2
import torch
import numpy as np
from urllib import request
import matplotlib
matplotlib.use('TkAgg')
from matplotlib import pyplot as plt
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False
from pathlib import Path
from pyzjr.core.general import is_gray_image, is_pil, is_url, is_file, is_numpy, is_list
from pyzjr.visualize.color.colorspace import bgr2rgb
from pyzjr.augmentation.PIC import convert_cv_to_pil

IMREAD_GRAYSCALE = cv2.IMREAD_GRAYSCALE
IMREAD_COLOR = cv2.IMREAD_COLOR
RGB2BGR  = cv2.COLOR_RGB2BGR
BGR2RGB  = cv2.COLOR_BGR2RGB
BGR2GRAY = cv2.COLOR_BGR2GRAY
RGB2GRAY = cv2.COLOR_RGB2GRAY
GRAY2BGR = cv2.COLOR_GRAY2BGR
GRAY2RGB = cv2.COLOR_GRAY2RGB
BGR2HSV  = cv2.COLOR_BGR2HSV
RGB2HSV  = cv2.COLOR_RGB2HSV
HSV2BGR  = cv2.COLOR_HSV2BGR
HSV2RGB  = cv2.COLOR_HSV2RGB

def display(winname, imgArray, scale=1.):
    """Displays an image in the specified window."""
    _imshow = cv2.imshow  # copy to avoid recursion errors
    if is_numpy(imgArray):
        height, width = imgArray.shape[:2]
        new_width = int(width * scale)
        new_height = int(height * scale)
        image = cv2.resize(imgArray, (new_width, new_height))
        _imshow(winname.encode('unicode_escape').decode(), image)
    elif is_list(imgArray):
        image = StackedImagesV1(scale, imgArray)
        _imshow(winname.encode('unicode_escape').decode(), image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()

def imreader(filename, is_gray=False, flip=None, to_rgb=False):
    """
    :param filename: Path to the image file.
    :param is_gray: If True, read the image in grayscale.
    :param flip: 0: Flip vertically.
                1: Flip horizontally.
               -1: Flip both vertically and horizontally.
    :param to_rgb: Convert to RGB, default to False
    """
    if not is_file(filename):
        raise ValueError(f"imreader: The file {filename} does not exist.")
    if isinstance(filename, Path):
        filename = str(filename.resolve())
    flags = IMREAD_GRAYSCALE if is_gray else IMREAD_COLOR
    image = cv2.imdecode(np.fromfile(filename, np.uint8), flags)
    if flip is not None:
        assert flip in [-1, 0, 1], f"imreader: The 'flip' parameter must be -1, 0, or 1."
        image = cv2.flip(image, flip)
    if is_gray_image(image):
        image = cv2.cvtColor(image, GRAY2RGB) if to_rgb else image
    else:
        image = bgr2rgb(image) if to_rgb else image
    return image

def imwrite(filename: str, img: np.ndarray, params=None):
    """Write the image to a file."""
    try:
        cv2.imencode(Path(filename).suffix, img, params)[1].tofile(filename)
        return True
    except Exception:
        return False

def url2image(url):
    if is_url(url):
        res = request.urlopen(url, timeout=3)
    else:
        raise ValueError("url2image: The current input parameter does not conform to the URL format")
    try:
        image = np.asarray(bytearray(res.read()), dtype="uint8")
        image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
    except:
        print('[url2imgcv2]: Load read - Image timeout!')
        image = []
    h, w, c = image.shape
    if c == 4:
        image = image[:, :, :3]
    return image

def improperties(image):
    if is_numpy(image):
        imshape = image.shape
        size = imshape[:2]
        dtype = image.dtype
        height, width = size

        if len(imshape) == 2:
            depth = image.itemsize * 8
        else:
            depth = image.itemsize * 8 * imshape[2]

        return {
            "shape": imshape,
            "size": size,
            "height": height,
            "width": width,
            "dtype": dtype,
            "depth": depth,
            "source": "NumPy"
        }
    elif is_pil(image):
        width, height = image.size
        mode = image.mode

        if mode == 'L':  # Grayscale
            depth = 8
        elif mode in ('RGB', 'RGBA'):
            depth = 8 * 3 if mode == 'RGB' else 8 * 4
        elif mode in ('I', 'F'):  # Integer or float (single channel)
            depth = image.getbands()[0].size * 8
        else:
            depth = "Unknown"  # Other modes are not handled here

        return {
            "shape": (height, width, len(image.getbands())),  # Simulate a NumPy shape
            "size": (width, height),
            "height": height,
            "width": width,
            "dtype": "PIL mode " + mode,  # PIL doesn't have a direct dtype equivalent
            "depth": depth,
            "source": "PIL"
        }
    else:
        return "Input is not a NumPy array or PIL Image."

def imshowplt(image, stacked_shape=None, title=None, cmap=None, save_path=None):
    """显示图像并添加标题"""
    import matplotlib
    matplotlib.use("TkAgg")
    if is_list(image):
        image = Stackedplt(image, stacked_shape)
    else:
        if is_pil(image):
            if image.mode == 'L' or (image.mode == 'LA' and image.getbands()[1] == 'L'):
                cmap = 'gray'
        elif is_gray_image(image):
            cmap = 'gray'
    plt.imshow(image, cmap=cmap)
    plt.axis('off')
    if title:
        plt.title(title)
    if save_path:
        plt.savefig(save_path)
    plt.show()

def StackedImagesV1(scale, imgArray):
    """
    :param scale:图片的规模, 可使用小数, 1为原始图像
    :param imgArray: 图像组成的列表，来表示行列排列
    :return: 生成的图像可按输入列表的行列排序展示, 拼接展示图像
    """
    rows = len(imgArray)
    cols = len(imgArray[0])
    rowsAvailable = isinstance(imgArray[0], list)
    width = imgArray[0][0].shape[1]
    height = imgArray[0][0].shape[0]
    if rowsAvailable:
        for x in range(0, rows):
            for y in range(0, cols):
                if imgArray[x][y].shape[:2] == imgArray[0][0].shape [:2]:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (0, 0), None, scale, scale)
                else:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (imgArray[0][0].shape[1], imgArray[0][0].shape[0]), None, scale, scale)
                if len(imgArray[x][y].shape) == 2: imgArray[x][y]= cv2.cvtColor( imgArray[x][y], cv2.COLOR_GRAY2BGR)
        imageBlank = np.zeros((height, width, 3), np.uint8)
        hor = [imageBlank] * rows
        for x in range(0, rows):
            hor[x] = np.hstack(imgArray[x])
        ver = np.vstack(hor)
    else:
        for x in range(0, rows):
            if imgArray[x].shape[:2] == imgArray[0].shape[:2]:
                imgArray[x] = cv2.resize(imgArray[x], (0, 0), None, scale, scale)
            else:
                imgArray[x] = cv2.resize(imgArray[x], (imgArray[0].shape[1], imgArray[0].shape[0]), None,scale, scale)
            if len(imgArray[x].shape) == 2:
                imgArray[x] = cv2.cvtColor(imgArray[x], cv2.COLOR_GRAY2BGR)
        hor = np.hstack(imgArray)
        ver = hor
    return ver

def StackedImagesV2(scale, imgList, cols):
    """
    将多张图像堆叠在一起并在单个窗口中显示
    :param scale: 图像的缩放比例,大于1表示放大,小于1表示缩小
    :param imgList: 要堆叠的图像列表
    :param cols: 每行显示的图像数量
    :return: 堆叠后的图像
    """
    totalImages = len(imgList)
    rows = totalImages // cols if totalImages // cols * cols == totalImages else totalImages // cols + 1
    blankImages = cols * rows - totalImages

    width = imgList[0].shape[1]
    height = imgList[0].shape[0]
    imgBlank = np.zeros((height, width, 3), np.uint8)
    imgList.extend([imgBlank] * blankImages)
    for i in range(cols * rows):
        imgList[i] = cv2.resize(imgList[i], (0, 0), None, scale, scale)
        if len(imgList[i].shape) == 2:
            imgList[i] = cv2.cvtColor(imgList[i], cv2.COLOR_GRAY2BGR)
    hor = [imgBlank] * rows
    for y in range(rows):
        line = []
        for x in range(cols):
            line.append(imgList[y * cols + x])
        hor[y] = np.hstack(line)
    ver = np.vstack(hor)
    return ver

def Stackedplt(images, shape):
    """
    按照指定形状堆叠多张图片
    :param images: 包含多张PIL Image对象的列表
    :param shape: 展示的图片的行列数,如(2, 3)
    :return: 堆叠后的PIL Image对象
    """
    rows, cols = shape
    new_images = []
    for im in images:
        if is_pil(im):
            new_images.append(im)
        elif is_numpy(im):
            new_images.append(convert_cv_to_pil(im))
    assert all(img.size == new_images[0].size for img in new_images), "All images must be the same size"
    width, height = new_images[0].size
    new_img_width = width * cols
    new_img_height = height * rows
    new_img = Image.new('RGB', (new_img_width, new_img_height), color=(255, 255, 255))
    x_offset = 0
    y_offset = 0
    for i in range(rows):
        for j in range(cols):
            index = i * cols + j
            if index < len(new_images):
                img = new_images[index]
                new_img.paste(img, (x_offset, y_offset))
                x_offset += width
        x_offset = 0
        y_offset += height

    return new_img

def Stackedtorch(imgs, num_rows, num_cols, titles=None, scale=None, camp=None):
    """
    堆叠显示Tensor图像或PIL图像
    :param imgs: 包含图像的列表,可以包含Tensor图像或PIL图像
    :param num_rows: 图像的行数
    :param num_cols: 图像的列数
    :param titles: 图像的标题列表,默认为None
    :param scale: 图像的缩放比例,默认为None
    :param camp: 'gray'
    :return:
    example:
        imgs = [image1, image2, image3, image4]
        titles = ['Image 1', 'Image 2', 'Image 3','Image 4']
        axes = Stackedtorch(imgs, 2, 2, titles=titles, scale=5)
        plt.savefig('my_plot.jpg')
        plt.show()
    """
    if scale:
        figsize = (num_cols * scale, num_rows * scale)
        _, axes = plt.subplots(num_rows, num_cols, figsize=figsize)
    else:
        _, axes = plt.subplots(num_rows, num_cols)
    axes = axes.flatten()
    for i, (ax, img) in enumerate(zip(axes, imgs)):
        if torch.is_tensor(img):
            # Tensor Image
            ax.imshow(img.numpy(), cmap=camp)
        else:
            # PIL Image
            ax.imshow(img, cmap=camp)
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)
        if titles:
            ax.set_title(titles[i])

    return axes

def plot_line(x, y, xlabel, ylabel, title):
    """
    绘制折线图
    :param x: x轴的数据列表
    :param y: y轴的数据列表
    :param xlabel: x轴的标签
    :param ylabel: y轴的标签
    :param title: 图形的标题
    :return:
    """
    plt.plot(x, y)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.show()

def bar_chart(x, y, xlabel, ylabel, title):
    """绘制柱状图"""
    plt.bar(x, y)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.show()

def scatter_plot(x, y, xlabel, ylabel, title):
    """绘制散点图"""
    plt.scatter(x, y)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.show()


if __name__=="__main__":
    from PIL import Image
    x = [1, 2, 3, 4, 5]
    y = [2, 4, 6, 8, 10]
    xlabel = 'X'
    ylabel = 'Y'
    title = '折线图示例'
    # plot_line(x, y, xlabel, ylabel, title)
    # bar_chart(x, y, xlabel, ylabel, title)
    # scatter_plot(x, y, xlabel, ylabel, title)

    image_path = r'D:\PythonProject\pyzjrPyPi\pyzjr\augmentation\test.png'
    # image = imreader(image_path)
    # gray_image = imreader(image_path, is_gray=True)
    # print(improperties(image), '\n', improperties(gray_image))
    # display("image-test", [[image, image], [gray_image, gray_image]], 1)
    # imshowplt([gray_image, image, gray_image, gray_image], (2,2))
    # imwrite("ss.png", image)

    image_pil = Image.open(image_path)
    print(improperties(image_pil))
    # imagess = Stackedtorch([image_pil, image_pil], 1, 2)
    import matplotlib
    matplotlib.use("TkAgg")
    axes = Stackedtorch([image_pil, image_pil], 1, 2, scale=5)
    plt.show()


from .drop import (
    DropPath,
    Dropout,
    MultiSampleDropout,
    DropConnect,
    Standout,
    GaussianDropout
)

from .Initer import (
    init_weights_complex,
    init_weights_simply,
    official_init,
    trunc_normal_,
    initialize_decoder,
    initialize_head
)

from .classfier import ClassifierHead, create_classifier

from .conv_norm_act import (
    ConvNormAct,
    ConvNorm,
    NormAct,
    conv3x3,
    conv1x1,
    ConvBnReLU,
    ConvBn,
    BnReLU
)

from .act import (
    ArgMax,
    Clamp,
    Activation
)

from .comblock import *


import torch
import torch.nn as nn
import numpy as np
from pyzjr.nn.Metrics.indexutils import do_metric_reduction_3d, hd, ignore_background
from pyzjr.core.general import is_tensor

class DiceMetric3d(nn.Module):
    def __init__(
            self,
            include_background=True,
            reduction='mean',
            get_not_nans=False,
            ignore_empty=True,
            num_classes=None,
            softmax=False,
            sigmoid=False
    ):
        super(DiceMetric3d, self).__init__()
        self.include_background = include_background
        self.reduction = reduction
        self.get_not_nans = get_not_nans
        self.ignore_empty = ignore_empty
        self.num_classes = num_classes
        self.softmax = softmax
        self.sigmoid = sigmoid

    def compute_channel(self, y_pred, y):
        y_o = torch.sum(y)
        if y_o > 0:
            return (2.0 * torch.sum(torch.masked_select(y, y_pred))) / (y_o + torch.sum(y_pred))
        if self.ignore_empty:
            return torch.tensor(float("nan"), device=y_o.device)
        denorm = y_o + torch.sum(y_pred)
        if denorm <= 0:
            return torch.tensor(1.0, device=y_o.device)
        return torch.tensor(0.0, device=y_o.device)

    def forward(self, y_pred, y):
        """
        Args:
            y_pred: input predictions with shape (batch_size, num_classes or 1, spatial_dims...).
                the number of channels is inferred from ``y_pred.shape[1]`` when ``num_classes is None``.
            y: ground truth with shape (batch_size, num_classes or 1, spatial_dims...).
        """
        _softmax, _sigmoid = self.softmax, self.sigmoid
        if self.num_classes is None:
            n_pred_ch = y_pred.shape[1]  # y_pred is in one-hot format or multi-channel scores
        else:
            n_pred_ch = self.num_classes
            if y_pred.shape[1] == 1 and self.num_classes > 1:  # y_pred is single-channel class indices
                _softmax = _sigmoid = False

        if _softmax:
            if n_pred_ch > 1:
                y_pred = torch.argmax(y_pred, dim=1, keepdim=True)

        elif _sigmoid:
            if self.activate:
                y_pred = torch.sigmoid(y_pred)
            y_pred = y_pred > 0.5

        first_ch = 0 if self.include_background else 1
        data = []
        for b in range(y_pred.shape[0]):
            c_list = []
            for c in range(first_ch, n_pred_ch) if n_pred_ch > 1 else [1]:
                x_pred = (y_pred[b, 0] == c) if (y_pred.shape[1] == 1) else y_pred[b, c].bool()
                x = (y[b, 0] == c) if (y.shape[1] == 1) else y[b, c]
                c_list.append(self.compute_channel(x_pred, x))
            data.append(torch.stack(c_list))
        data = torch.stack(data, dim=0).contiguous()

        f, not_nans = do_metric_reduction_3d(data, self.reduction)  # type: ignore
        return (f, not_nans) if self.get_not_nans else f

class HausdorffDistanceMetric3d(nn.Module):
    def __init__(
            self,
            include_background=True,
            percentile=None,
    ):
        super(HausdorffDistanceMetric3d, self).__init__()
        self.include_background = include_background
        self.percentile = percentile

    def forward(self, y_pred, y):
        if not self.include_background:
            y_pred, y = ignore_background(y_pred=y_pred, y=y)
        if y.shape != y_pred.shape:
            raise ValueError(f"y_pred and y should have same shapes, got {y_pred.shape} and {y.shape}.")

        batch_size, n_class = y_pred.shape[:2]
        hd_ = []
        if is_tensor(y_pred) and is_tensor(y):
            y_pred, y = y_pred.numpy(), y.numpy()

        for c in np.ndindex(n_class):
            surface_distance = hd(y_pred[:, c,...], y[:, c,...])
            hd_.append(surface_distance)

        if 0 <= self.percentile <= 100:
            return np.percentile(hd_, self.percentile)

class MeanIoUMetric3d(nn.Module):
    def __init__(
            self,
            include_background=True,
            reduction='mean',
            get_not_nans=False,
            ignore_empty=True,
    ):
        super().__init__()
        self.include_background = include_background
        self.reduction = reduction
        self.get_not_nans = get_not_nans
        self.ignore_empty = ignore_empty

    def forward(self, y_pred, y):
        if not self.include_background:
            y_pred, y = ignore_background(y_pred=y_pred, y=y)

        y = y.float()
        y_pred = y_pred.float()

        if y.shape != y_pred.shape:
            raise ValueError(f"y_pred and y should have same shapes, got {y_pred.shape} and {y.shape}.")

        # reducing only spatial dimensions (not batch nor channels)
        n_len = len(y_pred.shape)
        reduce_axis = list(range(2, n_len))
        intersection = torch.sum(y * y_pred, dim=reduce_axis)

        y_o = torch.sum(y, reduce_axis)
        y_pred_o = torch.sum(y_pred, dim=reduce_axis)
        union = y_o + y_pred_o - intersection

        if self.ignore_empty:
            dice = torch.where(y_o > 0, (intersection) / union, torch.tensor(float("nan"), device=y_o.device))
        else:
            dice = torch.where(union > 0, (intersection) / union, torch.tensor(1.0, device=y_o.device))

        f, not_nans = do_metric_reduction_3d(dice, self.reduction)  # type: ignore
        return (f, not_nans) if self.get_not_nans else f

if __name__=="__main__":
    batch_size = 1
    num_classes = 3
    depth = 5
    height = 10
    width = 10
    y_pred = torch.rand(batch_size, num_classes, depth, height, width)
    y = torch.randint(0, num_classes, (batch_size, num_classes, depth, height, width))
    dice_metric = DiceMetric3d(include_background=True, reduction="none")
    hd_metric = HausdorffDistanceMetric3d(include_background=True, percentile=95)
    iou_metric = MeanIoUMetric3d(include_background=True, reduction="none")
    dice_score = dice_metric(y_pred, y)
    print("Dice Score:", dice_score)
    hd_score = hd_metric(y_pred, y)
    print("HD Score:", hd_score)
    iou_score = iou_metric(y_pred, y)
    print("IOU Score:", iou_score)


import yaml
from pyzjr.core.general import is_not_None

def yamlread(path, name=None):
    with open(path) as file:
        data = yaml.load(file, Loader=yaml.FullLoader)
    value = data[name] if is_not_None(name) else data

    return value


import re
import os
import torch
import torch.nn as nn
import numpy as np
from PIL import Image
import pyzjr.Z as Z
import platform

def is_tensor(img):
    """Check if the input image is torch format."""
    return isinstance(img, torch.Tensor)

def is_pil(img):
    """Check if the input image is PIL format."""
    return isinstance(img, Image.Image)

def is_numpy(img):
    """Check if the input image is Numpy format."""
    return isinstance(img, np.ndarray)

def is_gray_image(image):
    return (len(image.shape) == 2) or (len(image.shape) == 3 and image.shape[-1] == 1)

def is_rgb_image(image):
    return len(image.shape) == 3 and image.shape[-1] == 3

def is_parallel(model):
    # Returns True if model is of type DP or DDP
    return isinstance(model, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel))

def is_str(param):
    return isinstance(param, str)

def is_int(param):
    return isinstance(param, int)

def is_float(param):
    return isinstance(param, float)

def is_bool(param):
    return param.dtype == np.bool_

def is_list(param):
    return isinstance(param, list)

def is_tuple(param):
    return isinstance(param, tuple)

def is_list_or_tuple(param):
    return isinstance(param, (list, tuple))

def is_None(param):
    return True if param is None else False

def is_not_None(param):
    return not is_None(param)

def is_positive_int(param):
    return is_int(param) and param > 0

def is_nonnegative_int(param):
    return is_int(param) and param >= 0

def is_ascii(s):
    """
    Check if the string is composed of only ASCII characters.
    Args:
        s (str): String to be checked.
    Returns:
        bool: True if the string is composed only of ASCII characters, False otherwise.
    """
    s = str(s)
    return all(ord(c) < 128 for c in s)

def is_url(filename):
    """Return True if string is an http or ftp path."""
    URL_REGEX = re.compile(r'http://|https://|ftp://|file://|file:\\')
    return (isinstance(filename, str) and
            URL_REGEX.match(filename) is not None)

def is_image_extension(image_name):
    ext = image_name.split('.')
    return ext[-1] in Z.IMG_FORMATS and is_str(image_name)

def is_video_extension(video_name):
    ext = video_name.split('.')
    return ext[-1] in Z.VID_FORMATS and is_str(video_name)

def is_file(path):
    return os.path.isfile(path) and is_str(path)

def is_directory(path):
    return os.path.isdir(path) and is_str(path)

def is_directory_not_empty(path):
    return os.path.isdir(path) and len(os.listdir(path)) > 0 and is_str(path)

def is_path_exists(path):
    return os.path.exists(path) and is_str(path)

def is_Windows():
    if platform.system() == "Windows":
        return True

def is_Linux():
    if platform.system() == "Linux":
        return True

if __name__=="__main__":
    image_name = r'D:\PythonProject\pyzjrPyPi\pyzjr\augmentation\test.png'
    print(is_file(image_name), is_Windows())


# 像素级别的研究

from .pixel import (
    SkeletonMap,
    incircle,
    outcircle,
    foreground_contour_length,
    get_each_crack_areas
)
from .steger import (
    Steger,
    _derivation_with_Filter,
    _derivation_with_Scharr,
    _derivation_with_Sobel,
    Magnitudefield,
    derivation,
    nonMaxSuppression,
)

from .crack import *
from .dehaze import *


"""
Example:

    .. code-block:: python


        # lets assume we have multilabel prediction for 3 classes
        output = torch.rand([10, 3, 256, 256])
        target = torch.rand([10, 3, 256, 256]).round().long()

        # first compute statistics for true positives, false positives, false negative and
        # true negative "pixels"
        tp, fp, fn, tn = get_stats(output, target, mode='multilabel', threshold=0.5)

        # then compute metrics with required reduction (see metric docs)
        iou_score = iou_score(tp, fp, fn, tn, reduction="micro")
        f1_score = f1_score(tp, fp, fn, tn, reduction="micro")
        f2_score = fbeta_score(tp, fp, fn, tn, beta=2, reduction="micro")
        accuracy = accuracy(tp, fp, fn, tn, reduction="macro")
        recall = recall(tp, fp, fn, tn, reduction="micro-imagewise")

segmentation_models_pytorch
"""
import torch
import warnings
from typing import Optional, List, Tuple, Union


__all__ = [
    "get_stats",
    "fbeta_score",
    "f1_score",
    "iou_score",
    "accuracy",
    "precision",
    "recall",
    "sensitivity",
    "specificity",
    "balanced_accuracy",
    "positive_predictive_value",
    "negative_predictive_value",
    "false_negative_rate",
    "false_positive_rate",
    "false_discovery_rate",
    "false_omission_rate",
    "positive_likelihood_ratio",
    "negative_likelihood_ratio",
]


###################################################################################################
# Statistics computation (true positives, false positives, false negatives, false positives)
###################################################################################################


def get_stats(
    output: Union[torch.LongTensor, torch.FloatTensor],
    target: torch.LongTensor,
    mode: str,
    ignore_index: Optional[int] = None,
    threshold: Optional[Union[float, List[float]]] = None,
    num_classes: Optional[int] = None,
) -> Tuple[torch.LongTensor, torch.LongTensor, torch.LongTensor, torch.LongTensor]:
    """Compute true positive, false positive, false negative, true negative 'pixels'
    for each image and each class.

    Args:
        output (Union[torch.LongTensor, torch.FloatTensor]): Model output with following
            shapes and types depending on the specified ``mode``:

            'binary'
                shape (N, 1, ...) and ``torch.LongTensor`` or ``torch.FloatTensor``

            'multilabel'
                shape (N, C, ...) and ``torch.LongTensor`` or ``torch.FloatTensor``

            'multiclass'
                shape (N, ...) and ``torch.LongTensor``

        target (torch.LongTensor): Targets with following shapes depending on the specified ``mode``:

            'binary'
                shape (N, 1, ...)

            'multilabel'
                shape (N, C, ...)

            'multiclass'
                shape (N, ...)

        mode (str): One of ``'binary'`` | ``'multilabel'`` | ``'multiclass'``
        ignore_index (Optional[int]): Label to ignore on for metric computation.
            **Not** supproted for ``'binary'`` and ``'multilabel'`` modes.  Defaults to None.
        threshold (Optional[float, List[float]]): Binarization threshold for
            ``output`` in case of ``'binary'`` or ``'multilabel'`` modes. Defaults to None.
        num_classes (Optional[int]): Number of classes, necessary attribute
            only for ``'multiclass'`` mode. Class values should be in range 0..(num_classes - 1).
            If ``ignore_index`` is specified it should be outside the classes range, e.g. ``-1`` or
            ``255``.

    Raises:
        ValueError: in case of misconfiguration.

    Returns:
        Tuple[torch.LongTensor]: true_positive, false_positive, false_negative,
            true_negative tensors (N, C) shape each.

    """

    if torch.is_floating_point(target):
        raise ValueError(f"Target should be one of the integer types, got {target.dtype}.")

    if torch.is_floating_point(output) and threshold is None:
        raise ValueError(
            f"Output should be one of the integer types if ``threshold`` is not None, got {output.dtype}."
        )

    if torch.is_floating_point(output) and mode == "multiclass":
        raise ValueError(f"For ``multiclass`` mode ``output`` should be one of the integer types, got {output.dtype}.")

    if mode not in {"binary", "multiclass", "multilabel"}:
        raise ValueError(f"``mode`` should be in ['binary', 'multiclass', 'multilabel'], got mode={mode}.")

    if mode == "multiclass" and threshold is not None:
        raise ValueError("``threshold`` parameter does not supported for this 'multiclass' mode")

    if output.shape != target.shape:
        raise ValueError(
            "Dimensions should match, but ``output`` shape is not equal to ``target`` "
            + f"shape, {output.shape} != {target.shape}"
        )

    if mode != "multiclass" and ignore_index is not None:
        raise ValueError(f"``ignore_index`` parameter is not supproted for '{mode}' mode")

    if mode == "multiclass" and num_classes is None:
        raise ValueError("``num_classes`` attribute should be not ``None`` for 'multiclass' mode.")

    if ignore_index is not None and 0 <= ignore_index <= num_classes - 1:
        raise ValueError(
            f"``ignore_index`` should be outside the class values range, but got class values in range "
            f"0..{num_classes - 1} and ``ignore_index={ignore_index}``. Hint: if you have ``ignore_index = 0``"
            f"consirder subtracting ``1`` from your target and model output to make ``ignore_index = -1``"
            f"and relevant class values started from ``0``."
        )

    if mode == "multiclass":
        tp, fp, fn, tn = _get_stats_multiclass(output, target, num_classes, ignore_index)
    else:
        if threshold is not None:
            output = torch.where(output >= threshold, 1, 0)
            target = torch.where(target >= threshold, 1, 0)
        tp, fp, fn, tn = _get_stats_multilabel(output, target)

    return tp, fp, fn, tn


@torch.no_grad()
def _get_stats_multiclass(
    output: torch.LongTensor,
    target: torch.LongTensor,
    num_classes: int,
    ignore_index: Optional[int],
) -> Tuple[torch.LongTensor, torch.LongTensor, torch.LongTensor, torch.LongTensor]:

    batch_size, *dims = output.shape
    num_elements = torch.prod(torch.tensor(dims)).long()

    if ignore_index is not None:
        ignore = target == ignore_index
        output = torch.where(ignore, -1, output)
        target = torch.where(ignore, -1, target)
        ignore_per_sample = ignore.view(batch_size, -1).sum(1)

    tp_count = torch.zeros(batch_size, num_classes, dtype=torch.long)
    fp_count = torch.zeros(batch_size, num_classes, dtype=torch.long)
    fn_count = torch.zeros(batch_size, num_classes, dtype=torch.long)
    tn_count = torch.zeros(batch_size, num_classes, dtype=torch.long)

    for i in range(batch_size):
        target_i = target[i]
        output_i = output[i]
        mask = output_i == target_i
        matched = torch.where(mask, target_i, -1)
        tp = torch.histc(matched.float(), bins=num_classes, min=0, max=num_classes - 1)
        fp = torch.histc(output_i.float(), bins=num_classes, min=0, max=num_classes - 1) - tp
        fn = torch.histc(target_i.float(), bins=num_classes, min=0, max=num_classes - 1) - tp
        tn = num_elements - tp - fp - fn
        if ignore_index is not None:
            tn = tn - ignore_per_sample[i]
        tp_count[i] = tp.long()
        fp_count[i] = fp.long()
        fn_count[i] = fn.long()
        tn_count[i] = tn.long()

    return tp_count, fp_count, fn_count, tn_count


@torch.no_grad()
def _get_stats_multilabel(
    output: torch.LongTensor,
    target: torch.LongTensor,
) -> Tuple[torch.LongTensor, torch.LongTensor, torch.LongTensor, torch.LongTensor]:

    batch_size, num_classes, *dims = target.shape
    output = output.view(batch_size, num_classes, -1)
    target = target.view(batch_size, num_classes, -1)

    tp = (output * target).sum(2)
    fp = output.sum(2) - tp
    fn = target.sum(2) - tp
    tn = torch.prod(torch.tensor(dims)) - (tp + fp + fn)

    return tp, fp, fn, tn


###################################################################################################
# Metrics computation
###################################################################################################


def _handle_zero_division(x, zero_division):
    nans = torch.isnan(x)
    if torch.any(nans) and zero_division == "warn":
        warnings.warn("Zero division in metric calculation!")
    value = zero_division if zero_division != "warn" else 0
    value = torch.tensor(value, dtype=x.dtype).to(x.device)
    x = torch.where(nans, value, x)
    return x


def _compute_metric(
    metric_fn,
    tp,
    fp,
    fn,
    tn,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division="warn",
    **metric_kwargs,
) -> float:

    if class_weights is None and reduction is not None and "weighted" in reduction:
        raise ValueError(f"Class weights should be provided for `{reduction}` reduction")

    class_weights = class_weights if class_weights is not None else 1.0
    class_weights = torch.tensor(class_weights).to(tp.device)
    class_weights = class_weights / class_weights.sum()

    if reduction == "micro":
        tp = tp.sum()
        fp = fp.sum()
        fn = fn.sum()
        tn = tn.sum()
        score = metric_fn(tp, fp, fn, tn, **metric_kwargs)

    elif reduction == "macro":
        tp = tp.sum(0)
        fp = fp.sum(0)
        fn = fn.sum(0)
        tn = tn.sum(0)
        score = metric_fn(tp, fp, fn, tn, **metric_kwargs)
        score = _handle_zero_division(score, zero_division)
        score = (score * class_weights).mean()

    elif reduction == "weighted":
        tp = tp.sum(0)
        fp = fp.sum(0)
        fn = fn.sum(0)
        tn = tn.sum(0)
        score = metric_fn(tp, fp, fn, tn, **metric_kwargs)
        score = _handle_zero_division(score, zero_division)
        score = (score * class_weights).sum()

    elif reduction == "micro-imagewise":
        tp = tp.sum(1)
        fp = fp.sum(1)
        fn = fn.sum(1)
        tn = tn.sum(1)
        score = metric_fn(tp, fp, fn, tn, **metric_kwargs)
        score = _handle_zero_division(score, zero_division)
        score = score.mean()

    elif reduction == "macro-imagewise" or reduction == "weighted-imagewise":
        score = metric_fn(tp, fp, fn, tn, **metric_kwargs)
        score = _handle_zero_division(score, zero_division)
        score = (score.mean(0) * class_weights).mean()

    elif reduction == "none" or reduction is None:
        score = metric_fn(tp, fp, fn, tn, **metric_kwargs)
        score = _handle_zero_division(score, zero_division)

    else:
        raise ValueError(
            "`reduction` should be in [micro, macro, weighted, micro-imagewise,"
            + "macro-imagesize, weighted-imagewise, none, None]"
        )

    return score


# Logic for metric computation, all metrics are with the same interface


def _fbeta_score(tp, fp, fn, tn, beta=1):
    beta_tp = (1 + beta**2) * tp
    beta_fn = (beta**2) * fn
    score = beta_tp / (beta_tp + beta_fn + fp)
    return score


def _iou_score(tp, fp, fn, tn):
    return tp / (tp + fp + fn)


def _accuracy(tp, fp, fn, tn):
    return (tp + tn) / (tp + fp + fn + tn)


def _sensitivity(tp, fp, fn, tn):
    return tp / (tp + fn)


def _specificity(tp, fp, fn, tn):
    return tn / (tn + fp)


def _balanced_accuracy(tp, fp, fn, tn):
    return (_sensitivity(tp, fp, fn, tn) + _specificity(tp, fp, fn, tn)) / 2


def _positive_predictive_value(tp, fp, fn, tn):
    return tp / (tp + fp)


def _negative_predictive_value(tp, fp, fn, tn):
    return tn / (tn + fn)


def _false_negative_rate(tp, fp, fn, tn):
    return fn / (fn + tp)


def _false_positive_rate(tp, fp, fn, tn):
    return fp / (fp + tn)


def _false_discovery_rate(tp, fp, fn, tn):
    return 1 - _positive_predictive_value(tp, fp, fn, tn)


def _false_omission_rate(tp, fp, fn, tn):
    return 1 - _negative_predictive_value(tp, fp, fn, tn)


def _positive_likelihood_ratio(tp, fp, fn, tn):
    return _sensitivity(tp, fp, fn, tn) / _false_positive_rate(tp, fp, fn, tn)


def _negative_likelihood_ratio(tp, fp, fn, tn):
    return _false_negative_rate(tp, fp, fn, tn) / _specificity(tp, fp, fn, tn)


def fbeta_score(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    beta: float = 1.0,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """F beta score"""
    return _compute_metric(
        _fbeta_score,
        tp,
        fp,
        fn,
        tn,
        beta=beta,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def f1_score(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """F1 score"""
    return _compute_metric(
        _fbeta_score,
        tp,
        fp,
        fn,
        tn,
        beta=1.0,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def iou_score(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """IoU score or Jaccard index"""  # noqa
    return _compute_metric(
        _iou_score,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def accuracy(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Accuracy"""
    return _compute_metric(
        _accuracy,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def sensitivity(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Sensitivity, recall, hit rate, or true positive rate (TPR)"""
    return _compute_metric(
        _sensitivity,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def specificity(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Specificity, selectivity or true negative rate (TNR)"""
    return _compute_metric(
        _specificity,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def balanced_accuracy(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Balanced accuracy"""
    return _compute_metric(
        _balanced_accuracy,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def positive_predictive_value(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Precision or positive predictive value (PPV)"""
    return _compute_metric(
        _positive_predictive_value,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def negative_predictive_value(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Negative predictive value (NPV)"""
    return _compute_metric(
        _negative_predictive_value,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def false_negative_rate(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Miss rate or false negative rate (FNR)"""
    return _compute_metric(
        _false_negative_rate,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def false_positive_rate(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Fall-out or false positive rate (FPR)"""
    return _compute_metric(
        _false_positive_rate,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def false_discovery_rate(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """False discovery rate (FDR)"""  # noqa
    return _compute_metric(
        _false_discovery_rate,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def false_omission_rate(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """False omission rate (FOR)"""  # noqa
    return _compute_metric(
        _false_omission_rate,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def positive_likelihood_ratio(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Positive likelihood ratio (LR+)"""
    return _compute_metric(
        _positive_likelihood_ratio,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


def negative_likelihood_ratio(
    tp: torch.LongTensor,
    fp: torch.LongTensor,
    fn: torch.LongTensor,
    tn: torch.LongTensor,
    reduction: Optional[str] = None,
    class_weights: Optional[List[float]] = None,
    zero_division: Union[str, float] = 1.0,
):
    """Negative likelihood ratio (LR-)"""
    return _compute_metric(
        _negative_likelihood_ratio,
        tp,
        fp,
        fn,
        tn,
        reduction=reduction,
        class_weights=class_weights,
        zero_division=zero_division,
    )


_doc = """

    Args:
        tp (torch.LongTensor): tensor of shape (N, C), true positive cases
        fp (torch.LongTensor): tensor of shape (N, C), false positive cases
        fn (torch.LongTensor): tensor of shape (N, C), false negative cases
        tn (torch.LongTensor): tensor of shape (N, C), true negative cases
        reduction (Optional[str]): Define how to aggregate metric between classes and images:

            - 'micro'
                Sum true positive, false positive, false negative and true negative pixels over
                all images and all classes and then compute score.

            - 'macro'
                Sum true positive, false positive, false negative and true negative pixels over
                all images for each label, then compute score for each label separately and average labels scores.
                This does not take label imbalance into account.

            - 'weighted'
                Sum true positive, false positive, false negative and true negative pixels over
                all images for each label, then compute score for each label separately and average
                weighted labels scores.

            - 'micro-imagewise'
                Sum true positive, false positive, false negative and true negative pixels for **each image**,
                then compute score for **each image** and average scores over dataset. All images contribute equally
                to final score, however takes into accout class imbalance for each image.

            - 'macro-imagewise'
                Compute score for each image and for each class on that image separately, then compute average score
                on each image over labels and average image scores over dataset. Does not take into account label
                imbalance on each image.

            - 'weighted-imagewise'
                Compute score for each image and for each class on that image separately, then compute weighted average
                score on each image over labels and average image scores over dataset.

            - 'none' or ``None``
                Same as ``'macro-imagewise'``, but without any reduction.

            For ``'binary'`` case ``'micro' = 'macro' = 'weighted'`` and
            ``'micro-imagewise' = 'macro-imagewise' = 'weighted-imagewise'``.

            Prefixes ``'micro'``, ``'macro'`` and ``'weighted'`` define how the scores for classes will be aggregated,
            while postfix ``'imagewise'`` defines how scores between the images will be aggregated.

        class_weights (Optional[List[float]]): list of class weights for metric
            aggregation, in case of `weighted*` reduction is chosen. Defaults to None.
        zero_division (Union[str, float]): Sets the value to return when there is a zero division,
            i.e. when all predictions and labels are negative. If set to “warn”, this acts as 0,
            but warnings are also raised. Defaults to 1.

    Returns:
        torch.Tensor: if ``'reduction'`` is not ``None`` or ``'none'`` returns scalar metric,
            else returns tensor of shape (N, C)

    References:
        https://en.wikipedia.org/wiki/Confusion_matrix
"""

fbeta_score.__doc__ += _doc
f1_score.__doc__ += _doc
iou_score.__doc__ += _doc
accuracy.__doc__ += _doc
sensitivity.__doc__ += _doc
specificity.__doc__ += _doc
balanced_accuracy.__doc__ += _doc
positive_predictive_value.__doc__ += _doc
negative_predictive_value.__doc__ += _doc
false_negative_rate.__doc__ += _doc
false_positive_rate.__doc__ += _doc
false_discovery_rate.__doc__ += _doc
false_omission_rate.__doc__ += _doc
positive_likelihood_ratio.__doc__ += _doc
negative_likelihood_ratio.__doc__ += _doc

precision = positive_predictive_value
recall = sensitivity

if __name__=="__main__":
    output = torch.rand([10, 3, 256, 256])
    target = torch.rand([10, 3, 256, 256]).round().long()

    # first compute statistics for true positives, false positives, false negative and
    # true negative "pixels"
    tp, fp, fn, tn = get_stats(output, target, mode='multilabel', threshold=0.5)

    # then compute metrics with required reduction (see metric docs)
    iou_score = iou_score(tp, fp, fn, tn, reduction="micro")
    f1_score = f1_score(tp, fp, fn, tn, reduction="micro")
    f2_score = fbeta_score(tp, fp, fn, tn, beta=2, reduction="micro")
    accuracy = accuracy(tp, fp, fn, tn, reduction="macro")
    recall = recall(tp, fp, fn, tn, reduction="micro-imagewise")
    print(iou_score, f1_score, f2_score, accuracy, recall)




from .general import (
    is_tensor,
    is_pil,
    is_numpy,
    is_gray_image,
    is_rgb_image,
    is_str,
    is_int,
    is_float,
    is_bool,
    is_list,
    is_tuple,
    is_list_or_tuple,
    is_None,
    is_not_None,
    is_positive_int,
    is_nonnegative_int,
    is_parallel,
    is_ascii,
    is_url,
    is_image_extension,
    is_video_extension,
    is_file,
    is_directory,
    is_directory_not_empty,
    is_path_exists,
    is_Linux,
    is_Windows
)

from ._utils import (
    get_num_channels,
    get_image_num_channels,
    get_image_size
)

from .decorator import timing
from .lr_scheduler import _LRScheduler
from .error import *

from .helpers import (
    _ntuple,
    to_1tuple,
    to_2tuple,
    to_3tuple,
    to_4tuple,
    to_ntuple,
    convert_to_tuple
)

from .__tensor import (
    to_tensor,
    hwc_and_chw,
    to_bchw,
    image_to_tensor,
    imagelist_to_tensor,
    tensor_to_image,
    img2tensor,
    label2tensor,
    SumExceptBatch
)


import cv2
import math
import numpy as np
import scipy.ndimage as ndi
from skimage.filters import sobel


def Brenner(img):
    '''
    Brenner梯度函数
    :param img:narray 二维灰度图像
    :return: int 图像越清晰越大
    '''
    shapes = np.shape(img)
    output = 0
    for x in range(0, shapes[0] - 2):
        for y in range(0, shapes[1]):
            output += (int(img[x + 2, y]) - int(img[x, y])) ** 2
    return output


def EOG(img):
    '''
    能量梯度函数(Energy of Gradient)
    :param img:narray 二维灰度图像
    :return: int 图像越清晰越大
    '''
    shapes = np.shape(img)
    output = 0
    for x in range(0, shapes[0] - 1):
        for y in range(0, shapes[1] - 1):
            output += ((int(img[x + 1, y]) - int(img[x, y])) ** 2 + (int(img[x, y + 1]) - int(img[x, y])) ** 2)
    return output

def Roberts(img):
    '''
    Roberts函数
    :param img:narray 二维灰度图像
    :return: int 图像越清晰越大
    '''
    shapes = np.shape(img)
    output = 0
    for x in range(0, shapes[0] - 1):
        for y in range(0, shapes[1] - 1):
            output += ((int(img[x + 1, y + 1]) - int(img[x, y])) ** 2 + (
                    int(img[x + 1, y]) - int(img[x, y + 1])) ** 2)
    return output

def Laplacian(img):
    '''
    Laplace
    :param img:narray 二维灰度图像
    :return: int 图像越清晰越大
    '''
    return cv2.Laplacian(img, cv2.Lap_64F).var()

def SMD(img):
    '''
    SMD(灰度方差)函数
    :param img:narray 二维灰度图像
    :return: int 图像越清晰越大
    '''
    shape = np.shape(img)
    output = 0
    for x in range(1, shape[0] - 1):
        for y in range(0, shape[1]):
            output += math.fabs(int(img[x, y]) - int(img[x, y - 1]))
            output += math.fabs(int(img[x, y] - int(img[x + 1, y])))
    return output

def SMD2(img):
    '''
    （灰度方差乘积）函数
    :param img:narray 二维灰度图像
    :return: int 图像约清晰越大
    '''
    shape = np.shape(img)
    output = 0
    for x in range(0, shape[0] - 1):
        for y in range(0, shape[1] - 1):
            output += math.fabs(int(img[x, y]) - int(img[x + 1, y])) * math.fabs(
                int(img[x, y] - int(img[x, y + 1])))
    return output

def Gradientmetric(img, size=11):
    """
    计算指示图像中模糊强度的度量(0表示无模糊, 1表示最大模糊)
    但实际上还是要根据我们的标准图与模糊图得到模糊区间
    :param img: 单通道进行处理,灰度图
    :param size: 重新模糊过滤器的大小
    :return: float, 唯一不是值越大, 越清晰的算法
    """
    image = np.array(img, dtype=np.float32) / 255
    n_axes = image.ndim
    shape = image.shape
    blur_metric = []

    slices = tuple([slice(2, s - 1) for s in shape])
    for ax in range(n_axes):
        filt_im = ndi.uniform_filter1d(image, size, axis=ax)
        im_sharp = np.abs(sobel(image, axis=ax))
        im_blur = np.abs(sobel(filt_im, axis=ax))
        T = np.maximum(0, im_sharp - im_blur)
        M1 = np.sum(im_sharp[slices])
        M2 = np.sum(T[slices])
        blur_metric.append(np.abs((M1 - M2)) / M1)

    return np.max(blur_metric) if len(blur_metric) > 0 else 0.0



from .opencv import (
    OpencvToTensor,
    OpencvResize,
    OpencvSquareResize,
    OpencvCenterzoom,
    OpencvHorizontalFlip,
    OpencvVerticalFlip,
    OpencvBrightness,
    OpencvAdjustGamma,
    OpencvToHSV,
    OpencvHistEqualize,
    OpencvRotation,
    OpencvLighting,
    OpencvRandomBlur,
    OpencvCrop,
    OpencvResizeCrop,
    OpencvPadResize,
    OpencvGrayscale,
)

from .pillow import (
    PILToTensor,
    NdarryToPIL,
    TensorToPIL,
    MeanStdNormalize,
    AdjustBrightness,
    AdjustContrast,
    AutoContrast,
    RandomAutoContrast,
    AdjustGamma,
    AdjustHue,
    AdjustSaturation,
    CenterCrop,
    EqualizeHistogram,
    RandomEqualizeHistogram,
    RandomHorizontalFlip,
    RandomVerticalFlip,
    RandomPCAnoise,
    InvertColor,
    RandomInvertColor,
    Resize,
    ColorJitter,
    RandomCrop,
    RandomRotation,
    Grayscale,
    RandomGrayscale,
    AdjustSharpness,
    RandomAdjustSharpness,
    GaussianBlur,
    ResizedCrop,
    RandomResizedCrop,
    Pad
)

from .tvision import (
    pad_if_smaller,
    ComposeWithLabel,
    RandomResize,
    RandomHorizontal_Flip,
    RandomVertical_Flip,
    Random_Crop,
    Center_Crop,
    ToTensor,
    Normalize,
    ToHSV,
    RandomContrast,
    RandomBrightness
)

from ._utils import (
    Images,
    random_apply,
    random_order,
    random_choice,
    uniform_augment,
    Compose,
    RandomApply,
    RandomChoice,
    RandomOrder,
    UniformAugment
)

from .normal import (
    imagenet_denormal,
    imagenet_normal,
    min_max_normal,
    z_score_normal,
    linear_normal,
    zero_centered_normal,
    Normalizer,
    denormalize
)



from .augments import (
    base_crop1,
    base_crop2,
    center_crop,
    five_crop,
    Stitcher_image,
    Centerzoom,
    flip,
    horizontal_flip,
    vertical_flip,
    resize,
    adjust_brightness_cv2,
    adjust_brightness_numpy,
    rotate,
    adjust_gamma,
    pad,
    erase,
    augment_Hsv,
    hist_equalize,
    random_resize_crop,
    random_crop,
    random_horizontal_flip,
    random_vertical_flip,
    random_rotation,
    random_lighting,
    Retinex
)

from .blur import (
    meanblur,
    medianblur,
    gaussianblur,
    bilateralblur,
    Filter,
)

from .PIC import (
    convert_cv_to_pil,
    convert_pil_to_cv,
    getContours,
    SearchOutline,
    drawOutline,
    gradientOutline,
    labelpoint,
    get_warp_perspective
)

from .mask_ops import (
    convert_np,
    clip,
    unique,
    RectMask,
    BinaryImg,
    up_low,
    approximate,
    ceilfloor,
    bool2mask,
    cv_distance,
    chamfer,
    fast_marching,
    addnoisy,
    inpaint_defect,
    cvt8png,
    cvtMask,
    imageContentMove,
    count_zero,
    count_white,
    count_nonzero
)

from .transforms import *


import cv2

import pyzjr.Z as Z
from pyzjr.data.reader import getPhotopath
from pyzjr.measure.dehaze.eval_definition import Laplacian, Gradientmetric

class Fuzzy_image():
    def __init__(self, mode="Laplacian"):
        """
        模糊判定区间: https://blog.csdn.net/m0_62919535/article/details/128061017
        :param mode: 图像清晰度算法模式,建议采用“Laplacian”或“Gradientmetric”算法,
                    其他的效果差
        """
        self.mode = mode
        if mode=='Laplacian':
            self.definemode = Laplacian
        elif mode=='Gradientmetric':
            self.definemode = Gradientmetric

    def getImgVar(self, image):
        """
        :param image: 图像
        :return: 图片清晰度数值
        """
        imggray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        imageVar = self.definemode(imggray)
        return abs(imageVar)

    def getReorder(self, imgfile, reverse=False):
        """
        :param imgfile: getPhotopath()获取图片路径的列表
        :param reverse: 表示是否进行反转排序(从大到小)
        :return: 模糊值排序
        """
        c = []
        for i in imgfile:
            img = cv2.imread(i)
            image = self.getImgVar(img)
            c.append(float(f"{image:.3f}"))
        c.sort(reverse=reverse)
        return c

    def getInterval(self, pathVague, pathStd):
        """
        :param pathVague:模糊图的数据集文件夹位置
        :param pathStd: 标准图的数据文件夹位置
        :return: 模糊判定区间
        """
        imgfile1, _ = getPhotopath(pathVague,debug=False)
        imgfile2, _ = getPhotopath(pathStd,debug=False)
        a = self.getReorder(imgfile1)
        b = self.getReorder(imgfile2)
        if self.mode == "Gradientmetric":
            thr = (b[-1], a[0])
        else:
            thr = (a[-1], b[0])
        return thr

def vagueJudge(img, pathVague, pathStd, mode="Laplacian", show_minandmax=True):
    """
    :param img: 图片
    :param show_minandmax:是否输出模糊未知区间
    :return: 模糊数值打印在图片上显示,只要没有超过清晰图像的阈值，全部判断为模糊
    """
    Fuzzy = Fuzzy_image(mode)
    imgVar = Fuzzy.getImgVar(img)
    minThr, maxThr = Fuzzy.getInterval(pathVague=pathVague, pathStd=pathStd)
    img2 = img.copy()
    if imgVar > maxThr:
        if mode == "G":
            text = "Vague"
            color = Z.red
        else:
            text = "Not Vague"
            color = Z.blue
    else:
        if mode == "G":
            text = "Not Vague"
            color = Z.blue
        else:
            text = "Vague"
            color = Z.red

    cv2.putText(img2, f"{text}{imgVar:.2f}", (12, 70), cv2.FONT_HERSHEY_PLAIN, 3, color, 3)

    if show_minandmax:
        print(minThr, maxThr)
    return img2

if __name__=="__main__":
    path = r"ces\Standards\001.jpg"   # 需要进行测试的图片
    path2 = r"ces\test\02.jpg"
    pathVague = r"ces\test"           # 模糊图像数据
    pathStd = r"ces\Standards"        # 标准图像数据
    img = cv2.imread(path)
    test = vagueJudge(img, pathVague, pathStd, mode="Laplacian")
    cv2.imshow("test", test)
    cv2.waitKey(0)


from .io import (
    imreader, display, imwrite, imshowplt, StackedImagesV1, StackedImagesV2, Stackedplt,
    Stackedtorch, plot_line, bar_chart, scatter_plot, url2image, improperties
)
from .color import *
from .mediapipe import Hands
from .video_utils import FPS, VideoCap, Timer, Mp4toGif
from .printf import set_logger, printlog, colorful, colorstr


import cv2
import numpy as np

def meanblur(img, ksize):
    """均值滤波 """
    blur_img = cv2.blur(img, ksize=ksize)
    return blur_img

def medianblur(img, ksize):
    """中值滤波"""
    if img.dtype == np.float32 and ksize not in {3, 5, 7, 9, 11}:
        raise ValueError(f"Invalid ksize value {ksize}.The available values are 3, 5, and 7")
    medblur_img = cv2.medianBlur(img, ksize=ksize)
    return medblur_img

def gaussianblur(img, ksize):
    """
    高斯模糊,提供给不熟悉高斯模糊参数的用户,sigma根据ksize进行自动计算,具体可以参考下面
    https://docs.opencv.org/master/d4/d13/tutorial_py_filtering.html
    """
    sigma = 0.3 * ((ksize - 1) * 0.5 - 1) + 0.8
    gaussianblur_img = cv2.GaussianBlur(img, ksize=(ksize, ksize), sigmaX=sigma)
    return gaussianblur_img

def bilateralblur(img, d=5, sigma_color=75, sigma_space=75):
    """双边滤波"""
    return cv2.bilateralFilter(img, d, sigma_color, sigma_space)

class Filter():
    """手写实现,仅供学习参考,最好还是使用cv2
    https://blog.csdn.net/m0_62919535/category_11936595.html?spm=1001.2014.3001.5482
    """
    def median_filtering(self, img,ksize=3):
        """
        中值滤波
        :param img:输入图像
        :param ksize: 核大小
        :return: 中值滤波平滑
        """
        h, w, c = img.shape
        half = ksize//2
        dst = np.zeros((h+2*half,w+2*half,c),np.uint8)
        dst[half:half+h, half:half+w] = img.copy()

        tmp=dst.copy()
        for y in range(h):
            for x in range(w):
                for z in range(c):
                    dst[half+x,half+y]=np.median(tmp[x:x+ksize,y:y+ksize])
        output=dst[half:half+h,half:half+w]
        return output

    def Arerage_Filtering(self, img, k_size=3):
        """
        均值滤波函数,默认会返回灰度图，因为三个for循环实在是太耗费时间了。而且，在这里需要考虑到边界点的问题。
        计算填充的宽度，即卷积核宽度的一半，用于处理图像边缘。使用cv2.copyMakeBorder函数进行边缘填充，将图
        像的边缘复制并填充到周围，以防止边缘像素点无法进行卷积。
        :param img: 原始图像，要求为灰度图
        :param k_size: 滤波核大小,默认为3,确保滤波核大小为奇数
        :return: 处理后的均值滤波图像
        """
        if k_size % 2 == 0:
            k_size += 1
        rows, cols = img.shape[:2]
        # 计算需要在图像边界扩充的大小
        pad_width = (k_size - 1) // 2
        # 在图像边界进行扩充
        img_pad = cv2.copyMakeBorder(img, pad_width, pad_width, pad_width, pad_width, cv2.BORDER_REPLICATE)
        img_filter = np.zeros_like(img)
        for i in range(rows):
            for j in range(cols):
                pixel_values = img_pad[i:i+k_size, j:j+k_size].flatten()
                img_filter[i, j] = np.mean(pixel_values)

        return img_filter

    def gaussian_kernel(self, size, sigma):
        """
        生成高斯核
        :param size: 核的大小
        :param sigma: 标准差
        :return:
        """
        kernel = np.zeros((size, size), dtype=np.float32)
        center = size // 2
        for i in range(size):
            for j in range(size):
                x, y = i - center, j - center
                kernel[i, j] = np.exp(-(x**2 + y**2)/(2*sigma**2))
        kernel /= 2 * np.pi * sigma**2
        kernel /= np.sum(kernel)
        return kernel

    def Gaussian_Filtering(self, img, kernel_size, sigma):
        """
        生成高斯滤波
        :param img: 输入图像
        :param kernel_size: 核大小
        :param sigma: 标准差
        :return:
        """
        kernel = self.gaussian_kernel(kernel_size, sigma)
        height, width, _ = img.shape
        result = np.zeros_like(img, dtype=np.float32)

        pad_size = kernel_size // 2
        img_pad = np.pad(img, [(pad_size, pad_size), (pad_size, pad_size), (0, 0)], mode='constant')
        for c in range(_):
            for i in range(pad_size, height + pad_size):
                for j in range(pad_size, width + pad_size):
                    result[i - pad_size, j - pad_size, c] = np.sum(kernel * img_pad[i - pad_size:i + pad_size + 1, j - pad_size:j + pad_size + 1, c])
        return np.uint8(result)


import torch
import torch.nn as nn
from pyzjr.nn.losses._utils import (
    sigmoid_focal_loss_3d,
    softmax_focal_loss_3d,
)
import warnings

class FocalLoss3D(nn.Module):
    """
    3D Focal Loss实现，用于处理3D图像分割任务。

    Args:
        include_background (bool): 是否包含背景类在损失计算中。默认为True。
        gamma (float): Focal Loss中的聚焦参数。默认为2.0。
        alpha (list or None): 用于平衡正负样本的权重。默认为None，表示不使用类别权重。
        use_softmax (bool): 是否使用softmax函数进行多分类。默认为False，表示使用sigmoid进行二分类或多标签分类。
        reduction (str): 指定损失函数的归约方式，可选值有'none', 'mean', 'sum'。默认为'mean'，表示计算损失的平均值。
    Example:
        >>> focal = FocalLoss3D()
        >>> batch_size = 2
        >>> num_classes = 3
        >>> d = h = w = 64
        >>> predictions = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
        >>> true_labels = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
        >>> loss = focal(predictions, true_labels)
        >>> print(f"Loss from FocalLoss3D: {loss.item()}")
    """
    def __init__(self,
                 include_background=True,
                 gamma=2.0,
                 alpha=None,
                 use_softmax=False,
                 reduction='mean'):
        super(FocalLoss3D, self).__init__()
        self.include_background = include_background
        self.gamma = gamma
        self.alpha = alpha
        self.use_softmax = use_softmax
        self.reduction = reduction

    def forward(self, input, target):
        n_pred_ch = input.shape[1]
        if not self.include_background:
            if n_pred_ch == 1:
                warnings.warn("single channel prediction, `include_background=False` ignored.")
            else:
                # if skipping background, removing first channel
                target = target[:, 1:]
                input = input[:, 1:]
        input = input.float()
        target = target.float()
        if target.shape != input.shape:
            raise ValueError(f"ground truth has different shape ({target.shape}) from input ({input.shape}),"
                             f"It may require one hot encoding")
        if self.use_softmax:
            if not self.include_background and self.alpha is not None:
                self.alpha = None
                warnings.warn("`include_background=False`, `alpha` ignored when using softmax.")
            loss = softmax_focal_loss_3d(input, target, self.gamma, self.alpha)
        else:
            loss = sigmoid_focal_loss_3d(input, target, self.gamma, self.alpha)

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')

class DiceLoss3D(nn.Module):
    """
    用于计算三维 Dice Loss 的 PyTorch 模块。
    注意“input”的轴N被期望为每个类的logits或概率，必须设置“sigmoid=True”或“softmax=True”
    Args:
        include_background (bool): 是否包括背景类，默认为 True。
        sigmoid (bool): 是否应用 sigmoid 函数，默认为 False。
        softmax (bool): 是否应用 softmax 函数，默认为 False。
        squared_pred (bool): 是否使用预测值的平方，默认为 False。
        jaccard (bool): 是否使用 Jaccard 损失（相当于 Dice Loss 的变体），默认为 False。
        reduction (str): 损失值的减少方式，可选值包括 "mean"、"sum" 和 "none"，默认为 "mean"。
        smooth_nr (float): 分子平滑参数，默认为 1e-5。
        smooth_dr (float): 分母平滑参数，默认为 1e-5。
    Example:
        >>> dice = DiceLoss3D()
        >>> batch_size = 2
        >>> num_classes = 3
        >>> d = h = w = 64
        >>> predictions = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
        >>> true_labels = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
        >>> loss = dice(predictions, true_labels)
        >>> print(f"Loss from DiceLoss3D: {loss.item()}")
    """
    def __init__(
            self,
            include_background=True,
            sigmoid=False,
            softmax=False,
            squared_pred=False,
            jaccard = False,
            reduction='mean',
            smooth_nr=1e-5,
            smooth_dr=1e-5,
    ):
        super(DiceLoss3D, self).__init__()
        self.include_background = include_background
        self.reduction = reduction
        self.sigmoid = sigmoid
        self.softmax = softmax
        self.squared_pred = squared_pred
        self.jaccard = jaccard
        self.smooth_nr = float(smooth_nr)
        self.smooth_dr = float(smooth_dr)

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if self.sigmoid:
            input = torch.sigmoid(input)

        n_pred_ch = input.shape[1]
        if self.softmax:
            if n_pred_ch == 1:
                warnings.warn("single channel prediction, `softmax=True` ignored.")
            else:
                input = torch.softmax(input, 1)

        if not self.include_background:
            if n_pred_ch == 1:
                warnings.warn("single channel prediction, `include_background=False` ignored.")
            else:
                target = target[:, 1:]
                input = input[:, 1:]

        if target.shape != input.shape:
            raise AssertionError(f"ground truth has different shape ({target.shape}) from input ({input.shape}),"
                                 f"It may require one hot encoding")

        reduce_axis = torch.arange(2, len(input.shape)).tolist()
        intersection = torch.sum(target * input, dim=reduce_axis)

        if self.squared_pred:
            ground_o = torch.sum(target**2, dim=reduce_axis)
            pred_o = torch.sum(input**2, dim=reduce_axis)
        else:
            ground_o = torch.sum(target, dim=reduce_axis)
            pred_o = torch.sum(input, dim=reduce_axis)

        denominator = ground_o + pred_o

        if self.jaccard:
            denominator = 2.0 * (denominator - intersection)

        f = 1.0 - (2.0 * intersection + self.smooth_nr) / (denominator + self.smooth_dr)

        if self.reduction == "mean":
            return torch.mean(f)
        elif self.reduction == "sum":
            return torch.sum(f)
        elif self.reduction == "none":
            return f
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')

class DiceFocalLoss3D(nn.Module):
    """
  用于结合 Dice Loss 和 Focal Loss 的三维损失函数模块。

    Args:
        include_background (bool): 是否包含背景类在损失计算中。默认为 True。
        sigmoid (bool): 是否应用 sigmoid 函数到预测值。默认为 False。
        softmax (bool): 是否应用 softmax 函数到预测值。默认为 False。多类别要改为 True。
        squared_pred (bool): 是否使用预测值的平方版本。仅用于 Dice Loss。
        jaccard (bool): 是否计算 Jaccard Index（软 IoU）而不是 Dice。默认为 False。
        reduction (str): 损失值的减少方式，可选值为 "mean"、"sum" 和 "none"。默认为 "mean"。
        smooth_nr (float): 分子平滑参数，默认为 1e-5。
        smooth_dr (float): 分母平滑参数，默认为 1e-5。
        gamma (float): Focal Loss 中的聚焦参数。默认为 2.0。
        lambda_dice (float): Dice Loss 的权重值。默认为 1.0。
        lambda_focal (float): Focal Loss 的权重值。默认为 1.0。
    Args:

    Example:
        >>> dice_focal = DiceFocalLoss3D()
        >>> batch_size = 2
        >>> num_classes = 3
        >>> d = h = w = 64
        >>> predictions = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
        >>> true_labels = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
        >>> loss = dice_focal(predictions, true_labels)
        >>> print(f"Loss from DiceFocalLoss3D: {loss.item()}")
    """
    def __init__(
            self,
            include_background=True,
            sigmoid=False,
            softmax=False,
            squared_pred=False,
            jaccard=False,
            reduction="mean",
            smooth_nr=1e-5,
            smooth_dr=1e-5,
            gamma=2.0,
            lambda_dice=1.0,
            lambda_focal=1.0,
    ):
        super(DiceFocalLoss3D, self).__init__()
        self.dice = DiceLoss3D(
            include_background=include_background,
            sigmoid=sigmoid,
            softmax=softmax,
            squared_pred=squared_pred,
            jaccard=jaccard,
            reduction='mean',
            smooth_nr=smooth_nr,
            smooth_dr=smooth_dr,
        )
        self.focal = FocalLoss3D(
            include_background=include_background,
            gamma=gamma,
            use_softmax=softmax,
            reduction='mean',
        )
        if lambda_dice < 0.0:
            raise ValueError("lambda_dice should be no less than 0.0.")
        if lambda_focal < 0.0:
            raise ValueError("lambda_focal should be no less than 0.0.")
        self.lambda_dice = lambda_dice
        self.lambda_focal = lambda_focal
        self.reduction = reduction

    def forward(self, input, target):
        dice_loss = self.dice(input, target)
        focal_loss = self.focal(input, target)
        total_loss = self.lambda_dice * dice_loss + self.lambda_focal * focal_loss

        if self.reduction == "mean":
            return torch.mean(total_loss)
        elif self.reduction == "sum":
            return torch.sum(total_loss)
        elif self.reduction == "none":
            return total_loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')


if __name__=="__main":
    focal = DiceFocalLoss3D()
    batch_size = 2
    num_classes = 3
    d = h = w = 64
    predictions = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
    true_labels = torch.randint(0, num_classes, (batch_size, num_classes, d, h, w))
    loss = focal(predictions, true_labels)
    print(f"Loss from FocalLoss3D: {loss.item()}")


# --------------------------------------- #
# Used for some operations on files,
# file management is abbreviated as FM.
# --------------------------------------- #
import os
import datetime
import csv
from datetime import datetime
from pathlib import Path
import logging

__all__=[
    "get_logger",
    "getFilePath",
    "mkdirs",
    "mkdir",
    "logdir",
    "Data2csv",
    "file_age",
    "file_date",
    "file_size",
]

def get_logger(filename='./exp.log', verbosity=1, name=None):
    """示例
    logger.debug('This is a DEBUG message.')  # DEBUG 级别的日志
    logger.info('This is an INFO message.')    # INFO 级别的日志
    logger.warning('This is a WARNING message.')  # WARNING 级别的日志
    try:
        result = 10 / 0
    except Exception as e:
        logger.error(f'An error occurred: {str(e)}', exc_info=True)
    logger.critical('This is a CRITICAL message.')  # CRITICAL 级别的日志
    """
    level_dict = {0: logging.DEBUG, 1: logging.INFO, 2: logging.WARNING}
    formatter = logging.Formatter(
        "[%(asctime)s][%(filename)s]%(message)s"
    )
    logger = logging.getLogger(name)
    logger.setLevel(level_dict[verbosity])

    fh = logging.FileHandler(filename, "w")
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    sh = logging.StreamHandler()
    sh.setFormatter(formatter)
    logger.addHandler(sh)

    return logger

def getFilePath(filedir, format='png'):
    """What is returned is a list that includes all paths under the target path that match the suffix."""
    file_list = [os.path.join(root, filespath) \
                 for root, dirs, files in os.walk(filedir) \
                 for filespath in files \
                 if str(filespath).endswith(format)
                 ]
    return file_list if file_list else []

def mkdirs(*args):
    for path in args:
        if not os.path.exists(path):
            os.makedirs(path)

def mkdir(path, *paths, inc=False):
    """
    如果构建的输出目录路径不存在，它将创建该目录。
    如果 inc 为 True，它将检查是否已经存在相同的目录。如果存在，将在目录名称末尾附加一个数字后缀，直到找到一个不存在的目录。
    返回最终的输出目录路径。
    """
    outdir = os.path.join(path, *paths)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    elif inc:
        count = 1
        outdir_inc = outdir + '_' + str(count)
        while os.path.exists(outdir_inc):
            count = count + 1
            outdir_inc = outdir + '_' + str(count)
            assert count < 100
        outdir = outdir_inc
        os.makedirs(outdir)
    return outdir


def logdir(dir="logs", format=True, prefix="", suffix=""):
    """
    Logging generator
    :param dir: Default "logs"
    :param format: DATE_FORMAT = '%A_%d_%B_%Y_%Hh_%Mm_%Ss'
                   SIMPLEDATE_FORMAT = '%Y_%m_%d_%H_%M_%S'
    :param prefix: Add folder name before time string
    :param suffix: After adding the folder name to the time string
    :return:
    """
    DATE_FORMAT = '%A_%d_%B_%Y_%Hh_%Mm_%Ss'
    SIMPLEDATE_FORMAT = '%Y_%m_%d_%H_%M_%S'
    formats = DATE_FORMAT if format else SIMPLEDATE_FORMAT
    time_str = datetime.now().strftime(formats)

    folder_names = [prefix, time_str, suffix]
    folder_names = [folder for folder in folder_names if folder]

    log_dir = os.path.join(dir, *folder_names)
    os.makedirs(log_dir, exist_ok=True)

    return log_dir

def Data2csv(header, value, log_dir, savefile_name):
    """Export data to CSV format
    Args:
        header (list): 列的标题
        value (list): 对应列的值
        log_dir (str): 文件夹路径
        savefile_name (str): 文件名（包括路径）
    """
    os.makedirs(log_dir, exist_ok=True)
    file_existence = os.path.isfile(savefile_name)

    if not file_existence:
        with open(savefile_name, 'w', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(header)
            writer.writerow(value)
    else:
        with open(savefile_name, 'a', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(value)

def file_age(path, detail=False):
    """Returns the number of days since the last file update."""
    dt = (datetime.now() - datetime.fromtimestamp(Path(path).stat().st_mtime))  # delta
    se = 0
    if detail:
        se = dt.seconds / 86400
    return f"{dt.days + se} days"

def file_date(path=__file__):
    """Return readable file modification date,i.e:'2021-3-26'."""
    t = datetime.fromtimestamp(Path(path).stat().st_mtime)
    return f'{t.year}-{t.month}-{t.day}'

def file_size(path):
    """Return file/dir size (MB)."""
    if isinstance(path, (str, Path)):
        mb = 1 << 20  # bytes -> MiB (1024 ** 2)
        path = Path(path)
        if path.is_file():
            size_mb = path.stat().st_size / mb
            return f"{size_mb:.5f} MB"
        elif path.is_dir():
            total_size = sum(f.stat().st_size for f in path.glob('**/*') if f.is_file())
            size_mb = total_size / mb
            return f"{size_mb:.5f} MB"
    return "0.0 MB"


from .segment2d import (
    Miou,
    Recall,
    Precision,
    F1Score,
    DiceCoefficient,
    Accuracy,
    SegmentationIndex,
    AIU
)

from .segment3d import (
    DiceMetric3d,
    HausdorffDistanceMetric3d,
    MeanIoUMetric3d
)

from .classification import (
    accuracy_all_classes,
    cls_matrix,
    BinaryConfusionMatrix,
    MulticlassConfusionMatrix,
    ConfusionMatrixs2D,
    ModelIndex,
    calculate_metrics,
    MultiLabelConfusionMatrix
)

from .indexutils import (
    hd, hd95,
    _surface_distances,
    ignore_background,
    _threshold,
    _take_channels,
    do_metric_reduction_3d
)

from .medical_index import (
    ConfusionMatrixs3D,
    get_confusion_matrix_3d,
    get_confusion_matrix_3d_np,
    get_confusion_matrix_3d_torch
)

from .functional import (
    get_stats,
    fbeta_score,
    f1_score,
    iou_score,
    accuracy,
    precision,
    recall,
    sensitivity,
    specificity,
    balanced_accuracy,
    positive_predictive_value,
    negative_predictive_value,
    false_negative_rate,
    false_positive_rate,
    false_discovery_rate,
    false_omission_rate,
    positive_likelihood_ratio,
    negative_likelihood_ratio,
)

from .infer_testset import (
    calculate_area,
    mean_iou,
    auc_roc,
    accuracy,
    dice,
    kappa,
    InferTestset
)


from pyzjr.core.general import is_numpy, is_tensor, is_pil, is_gray_image

def get_num_channels(image):
    return image.shape[2] if len(image.shape) == 3 else 1

def get_image_size(image):
    if is_numpy(image):
        h, w = image.shape[:2]
        return h, w

    if is_pil(image):
        w, h = image.size
        return h, w

    if is_tensor(image):
        if len(image.shape) == 4 or len(image.shape) == 3:
            w, h = image.shape[-2:]
            return h, w
    else:
        raise ValueError("Unsupported input type")

def get_image_num_channels(img):
    if is_tensor(img):
        if img.ndim == 2:
            return 1
        elif img.ndim > 2:
            return img.shape[-3]

    if is_pil(img):
        return 1 if img.mode == 'L' else 3

    if is_numpy(img):
        return 1 if is_gray_image else 3


import torch
import numpy as np
from scipy.ndimage import _ni_support, binary_erosion, distance_transform_edt, generate_binary_structure

def _surface_distances(y_pred, y, voxelspacing=None, connectivity=1):
    """
    计算两个二值对象中表面体素之间的距离
    """
    y_pred = np.atleast_1d(y_pred.astype(np.bool_))
    y = np.atleast_1d(y.astype(np.bool_))
    if voxelspacing is not None:
        voxelspacing = _ni_support._normalize_sequence(voxelspacing, y_pred.ndim)
        voxelspacing = np.asarray(voxelspacing, dtype=np.float64)
        if not voxelspacing.flags.contiguous:
            voxelspacing = voxelspacing.copy()

    # binary structure
    footprint = generate_binary_structure(y_pred.ndim, connectivity)

    # test for emptiness
    if 0 == np.count_nonzero(y_pred):
        raise RuntimeError(
            "The first supplied array does not contain any binary object."
        )
    if 0 == np.count_nonzero(y):
        raise RuntimeError(
            "The second supplied array does not contain any binary object."
        )
    result_border = y_pred ^ binary_erosion(y_pred, structure=footprint, iterations=1)
    reference_border = y ^ binary_erosion(
        y, structure=footprint, iterations=1
    )
    dt = distance_transform_edt(~reference_border, sampling=voxelspacing)
    sds = dt[result_border]

    return sds

def hd(y_pred, y, voxelspacing=None, connectivity=1):
    """
    Hausdorff Distance.
    Hausdorff距离衡量了两个点集之间的最大不匹配程度。在此函数中，它用于计算两个二值图像（result和reference）的表面体素之间的最大不匹配距离。

    Args:
        y_pred (array_like): 预测的二值图像，形状与参考图像（y）相同。
        y (array_like): 参考的二值图像，形状与预测图像（y_pred）相同。
        voxelspacing (array_like, optional): 形状为（N，）的数组，表示每个维度上的体素间距。默认为None。
        connectivity (int, optional): 表面距离计算中使用的连接性。默认为1，表示3D中的26连通（在3D情况下）或2D中的8连通（在2D情况下）。

    Returns:
        float: Hausdorff距离，即两个图像表面体素之间的最大不匹配距离。
    """
    hd1 = _surface_distances(y_pred, y, voxelspacing, connectivity).max()
    hd2 = _surface_distances(y, y_pred, voxelspacing, connectivity).max()
    hausdorff = max(hd1, hd2)
    return hausdorff

def hd95(y_pred, y, voxelspacing=None, connectivity=1):
    """
    计算 Hausdorff 距离的 95% 的数据点的值
    """
    hausdorff = hd(y_pred, y, voxelspacing, connectivity)
    return np.percentile(hausdorff, 95)

def ignore_background(y_pred, y):
    """
    Used to remove the background (first channel) of "yyred" and "y".
    Args:
        y_pred: predictions. As for classification tasks, `y_pred` should has the shape [BN] where N is larger than 1. As for segmentation tasks,
        the shape should be [BNHW] or [BNHWD].
        y: ground truth, the first dim is batch.
    """
    y = y[:, 1:] if y.shape[1] > 1 else y
    y_pred = y_pred[:, 1:] if y_pred.shape[1] > 1 else y_pred
    return y_pred, y

def _threshold(x, threshold=None):
    if threshold is not None:
        return (x > threshold).type(x.dtype)
    else:
        return x

def _take_channels(*xs, ignore_channels=None):
    if ignore_channels is None:
        return xs
    else:
        channels = [channel for channel in range(xs[0].shape[1]) if channel not in ignore_channels]
        xs = [torch.index_select(x, dim=1, index=torch.tensor(channels).to(x.device)) for x in xs]
        return xs

def do_metric_reduction_3d(f, reduction='mean'):
    """
    This function is to do the metric reduction for calculated `not-nan` metrics of each sample's each class.
    The function also returns `not_nans`, which counts the number of not nans for the metric.

    Args:
        f: a tensor that contains the calculated metric scores per batch and
            per class. The first two dims should be batch and class.
        reduction: define the mode to reduce metrics, will only apply reduction on `not-nan` values,
            available reduction modes: {``"none"``, ``"mean"``, ``"sum"``, ``"mean_batch"``, ``"sum_batch"``,
            ``"mean_channel"``, ``"sum_channel"``}, default to ``"mean"``.
            if "none", return the input f tensor and not_nans.

    Raises:
        ValueError: When ``reduction`` is not one of
            ["mean", "sum", "mean_batch", "sum_batch", "mean_channel", "sum_channel" "none"].
    """
    # some elements might be Nan (if ground truth y was missing (zeros))
    # we need to account for it
    nans = torch.isnan(f)
    not_nans = (~nans).float()

    t_zero = torch.zeros(1, device=f.device, dtype=f.dtype)
    if reduction == 'none':
        return f, not_nans

    f[nans] = 0
    if reduction == 'mean':
        # 2 steps, first, mean by channel (accounting for nans), then by batch
        not_nans = not_nans.sum(dim=1)
        f = torch.where(not_nans > 0, f.sum(dim=1) / not_nans, t_zero)  # channel average

        not_nans = (not_nans > 0).float().sum(dim=0)
        f = torch.where(not_nans > 0, f.sum(dim=0) / not_nans, t_zero)  # batch average

    elif reduction == 'sum':
        not_nans = not_nans.sum(dim=[0, 1])
        f = torch.sum(f, dim=[0, 1])  # sum over the batch and channel dims
    elif reduction == 'mean_batch':
        not_nans = not_nans.sum(dim=0)
        f = torch.where(not_nans > 0, f.sum(dim=0) / not_nans, t_zero)  # batch average
    elif reduction == 'sum_batch':
        not_nans = not_nans.sum(dim=0)
        f = f.sum(dim=0)  # the batch sum
    elif reduction == 'mean_channel':
        not_nans = not_nans.sum(dim=1)
        f = torch.where(not_nans > 0, f.sum(dim=1) / not_nans, t_zero)  # channel average
    elif reduction == 'sum_channel':
        not_nans = not_nans.sum(dim=1)
        f = f.sum(dim=1)  # the channel sum
    elif reduction != 'none':
        raise ValueError(
            f"Unsupported reduction: {reduction}, available options are "
            '["mean", "sum", "mean_batch", "sum_batch", "mean_channel", "sum_channel" "none"].'
        )
    return f, not_nans

if __name__=="__main__":
    y_pred = np.random.randint(0, 3, size=(100, 100))
    y = np.random.randint(0, 3, size=(100, 100))
    hausdorff_distance = hd95(y_pred, y)
    print("Hausdorff Distance:", hausdorff_distance)

    batch_size = 1
    num_channels = 3
    depth = 2
    height = 2
    width = 2

    y_pred = torch.rand(batch_size, num_channels, depth, height, width)
    y_true = torch.rand(batch_size, num_channels, depth, height, width)
    print("Shape of y_pred:", y_pred.shape)
    print("Shape of y_true:", y_true.shape)


import os
import torch
from pyzjr.core.general import is_not_None

_torch_save = torch.save  # copy to avoid recursion errors
best_val_loss = float('inf')

def save_model_to_pth_best(model, save_dir, val_loss, epoch, save_period=None):
    """
    Save the model based on training rounds (optional),
    and save the best model based on validation loss.

    Args:
        model: model to be saved
        save_dir: path that the model would be saved
        val_loss (float): Verification losses for the current period.
        epoch: the epoch the model finished training
        save_period (int, optional): The frequency of saving the model during training.
                                    The default is None.
    """
    global best_val_loss
    os.makedirs(save_dir, exist_ok=True)
    if epoch % save_period == 0 and is_not_None(save_period):
        _torch_save(model.state_dict(),
                   os.path.join(save_dir, f'model_epoch_{epoch}_loss_{val_loss:.4}.pth'))

    if val_loss < best_val_loss:
        best_val_loss = val_loss
        best_model_path = os.path.join(save_dir, "best_model.pth")
        _torch_save(model.state_dict(), best_model_path)
        print(f"Best model saved at epoch —— {epoch}")
        print(f'Save best model to {best_model_path}')


def save_model_to_pth_simplify(model, save_dir, epoch, save_period=10):
    """
    Save the model according to training rounds

    Args:
        model: model to be saved
        save_dir: path that the model would be saved
        epoch: the epoch the model finished training
        save_period: The frequency of saving the model during training.
                    The default is 10.
    """
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    if epoch % save_period == 0:
        _torch_save(model.state_dict(), os.path.join(save_dir, f"model_epoch_{epoch}.pth"))



import os
from PIL import Image
from pyzjr.data.basedataset import BaseDataset

__all__ = ["BaseClsDataset", "VOCSegmentation"]


class BaseClsDataset(BaseDataset):

    def __init__(self, file_txt, image_transform=None):
        """
        Supports txt file paths that match the format, or have been processed into the format list below.
        file_txt format: [path/to/xxx.jpg 0,
                          path/to/xxx.jpg 1,           OR           path/to/xxx.txt
                          path/to/xxx.jpg 2,
                          ...]
        """
        super().__init__()
        file_txt_result = self.load_txt_path_or_list(file_txt)
        if self._check_file_txt_format(file_txt_result):
            self.file_txt = file_txt_result
        self.image_transform = image_transform

    def __len__(self):
        return len(self.file_txt)

    def _check_file_txt_format(self, file_txt):
        for line in file_txt:
            parts = line.split()
            assert len(parts) == 2, f"BaseClsDataset: Invalid format in line: {line}"
            file_path, digit_label = parts
            assert os.path.exists(file_path), f"BaseClsDataset: File not found: {file_path}"
            assert digit_label.isdigit(), f"BaseClsDataset: Invalid digit label: {digit_label}"
        return True

    def __getitem__(self, idx):
        print(self.file_txt[idx])
        file_path, digit_label = self.file_txt[idx].split()
        raw_image = Image.open(file_path)
        image = raw_image.convert("RGB")
        if self.image_transform is not None:
            image = self.image_transform(image)
        return image, int(digit_label)

class VOCSegmentation(BaseDataset):
    def __init__(
            self,
            root,
            year="2012",
            image_set="train",
            transforms=None,
    ):
        super(VOCSegmentation, self).__init__()
        assert year in ["2007", "2008", "2009", "2010", "2011", "2012"], \
            "year can only choose 2007 to 2012"
        if year == "2007" and image_set == "test":
            year = "2007-test"
        self.year = year
        self.transforms = transforms
        voc_root = os.path.join(root, f"VOC{year}")
        txt_name = image_set + ".txt"

        assert os.path.exists(voc_root), "path '{}' does not exist.".format(voc_root)

        image_dir = os.path.join(voc_root, 'JPEGImages')
        mask_dir = os.path.join(voc_root, 'SegmentationClass')
        txt_path = os.path.join(voc_root, "ImageSets", "Segmentation", txt_name)

        assert os.path.exists(txt_path), "file '{}' does not exist.".format(txt_path)

        file_names = self.load_txt_path_or_list(txt_path)
        self.images = [os.path.join(image_dir, x + ".jpg") for x in file_names]
        self.masks = [os.path.join(mask_dir, x + ".png") for x in file_names]
        assert (len(self.images) == len(self.masks))

    def __getitem__(self, index: int):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is the image segmentation.
        """
        img = Image.open(self.images[index]).convert('RGB')
        target = Image.open(self.masks[index]).convert('L')

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target

    def __len__(self):
        return len(self.images)





# if __name__ == "__main__":
#     from torch.utils.data import DataLoader
#
#     dataset_path = r"D:\PythonProject\uhrnet\data\pneumonia"
#     train_dataset = SegmentTwoDataset(dataset_path, train=True)
#     val_dataset = SegmentTwoDataset(dataset_path, train=False)
#
#     train_loader = DataLoader(train_dataset, shuffle=True, batch_size=2, num_workers=8, pin_memory=True,
#                               drop_last=True)
#     val_loader = DataLoader(val_dataset, shuffle=True, batch_size=1, num_workers=4, pin_memory=True,
#                             drop_last=True)
#
#     for im, label, seglabel in train_loader:
#         print(im.shape, label.shape, seglabel.shape)


import torch
from torch.cuda.amp import GradScaler, autocast
from tqdm import tqdm
from pyzjr.devices import load_owned_device
from pyzjr.dlearn import get_lr
import logging
logging.basicConfig(level=logging.INFO, format='\033[34m%(message)s')
logger = logging.getLogger("loggers")

class SegmentationTrainEpoch():
    def __init__(self,
                 model,
                 loss_function,
                 optimizer,
                 lr_scheduler,
                 total_epoch,
                 use_amp=False,
                 device=load_owned_device()):
        super().__init__()
        self.device = device
        self.model = model.to(device)
        self.model_train = self.model.train()
        self.loss_function = loss_function
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.total_epoch = total_epoch
        self.scaler = None
        if use_amp:
            self.scaler = GradScaler()

    def train_one_epoch(self, train_loader, epoch):
        total_loss = 0
        total_batches = len(train_loader)
        logger.info('Start Train')
        with tqdm(total=self.total_epoch, desc=f'Epoch {epoch}/{self.total_epoch}', postfix=dict, mininterval=0.3) as pbar:
            for iteration, batch in enumerate(train_loader):
                if iteration >= self.total_epoch:
                    break
                imgs, pngs, labels = batch

                with torch.no_grad():
                    imgs = imgs.type(torch.FloatTensor)
                    pngs = pngs.long()
                    labels = labels.type(torch.FloatTensor)
                    if torch.cuda.is_available():
                        imgs = imgs.cuda()
                        pngs = pngs.cuda()
                        labels = labels.cuda()

                self.optimizer.zero_grad()
                if self.scaler is None:
                    outputs = self.model_train(imgs)
                    loss = self.loss_function(outputs, pngs, labels)
                    loss.backward()
                    self.optimizer.step()
                    self.lr_scheduler.step()
                else:
                    with autocast():
                        outputs = self.model_train(imgs)
                        loss = self.loss_function(outputs, pngs, labels)
                    self.scaler.scale(loss).backward()
                    self.scaler.step(self.optimizer)
                    self.scaler.update()
                total_loss += loss.item()

                pbar.set_postfix(**{'total_loss': total_loss / (iteration+1),
                                    'lr': get_lr(self.optimizer)})
                pbar.update(1)
        logger.info('Finish Train')
        return total_loss / total_batches

    def evaluate(self, val_loader, epoch):
        val_loss = 0
        self.model_train.eval()
        logger.info('Start Validation')
        total_batches = len(val_loader)
        with tqdm(total=self.total_epoch, desc=f'Epoch {epoch}/{self.total_epoch}', postfix=dict, mininterval=0.3) as pbar:
            for iteration, batch in enumerate(val_loader):
                if iteration >= self.total_epoch:
                    break
                imgs, pngs, labels = batch

                with torch.no_grad():
                    imgs = imgs.type(torch.FloatTensor)
                    pngs = pngs.long()
                    labels = labels.type(torch.FloatTensor)
                    if torch.cuda.is_available():
                        imgs = imgs.cuda()
                        pngs = pngs.cuda()
                        labels = labels.cuda()

                    outputs = self.model_train(imgs)
                    loss = self.loss_function(outputs, pngs, labels)
                    val_loss += loss.item()
                pbar.set_postfix(**{'total_loss': val_loss / (iteration+1)})
                pbar.update(1)

        logger.info('Finish Validation')
        return val_loss / total_batches




from .attribute import (
    crop_crack_according_to_bbox,
    crack_labels,
    DetectCrack,
)

from .skeleton_extraction import (
    skeletoncv,
    sketionio,
    sketion_medial_axis
)

from .crack_type import (
    CrackType,
    infertype,
    _get_minAreaRect_information
)